{------------------------------------------------------------------------------}
{                                                                              }
{ File:         captest.c                                                      }
{ RCS:          $Header: $                                                     }
{ Description:  Capability negotiation for Twacker                             }
{ Author:       TWAIN Working Group                                            }
{ Created:      Feb 5,92                                                       }
{ Modified:     "                                                              }
{ Language:     C                                                              }
{ Package:      N/A                                                            }
{ Status:       Test tool                                                      }
{                                                                              }
{ (c) Copyright 1992, Hewlett-Packard Company, all rights reserved.            }
{                                                                              }
{                                                                              }
{ This source code was converted from C to Delphi/Turbo Pascal for Windoes by  }
{                                                                              }
{   MCM Design                                                                 }
{   Marc Christopher Martin                                                    }
{   Australiensvej 34                                                          }
{   2100 Copenhagen OE                                                         }
{   Denmark                                                                    }
{                                                                              }
{   e-mail support@mcm-design.dk                                               }
{                                                                              }
{------------------------------------------------------------------------------}

{------------------------------------------------------------------------------}
{$A+} { Align data Byte = $A-, Word = $A+.                                     }
{$F+} { Force FAR call.                                                        }
{$G+} { Generate 80286 code.                                                   }
{$N-} { CoProcessor code generation.                                           }
{$R-} { Range check.                                                           }
{$W-} { Protected mode only $W-, Real mode stack frame support $W+.            }
{$X+} { Extended syntax (When on, function can be used as procedures).         }
{------------------------------------------------------------------------------}

Unit twa_capt;

{------------------------------------------------------------------------------}
{ MCM, General differences from original code.                                 }
{                                                                              }
{ (C - Twain) -> (pascal)                                                      }
{ sprintf     ->  StrFmt.                                                      }
{                                                                              }
{------------------------------------------------------------------------------}

interface

{------------------------------------------------------------------------------}
{                            I n c l u d e s                                   }
{------------------------------------------------------------------------------}

uses WinTypes,
     twain;

const
  CAP_AUTOFINAL = CAP_CUSTOMBASE + 1;

  ML_NONE 	= 0;
  ML_ERROR	= 1;
  ML_INFO	= 2;
  ML_FULL 	= 3;


{------------------------------------------------------------------------------}
{                            V a r i a b l e s                                 }
{------------------------------------------------------------------------------}

var UseManualFeedBOOL : integer; { = FALSE;                                    }
    AutoFeedBOOL      : bool;    { = FALSE;                                    }
    MessageLevelVAR   : integer; { = ERROR; To display what is happening       }


{------------------------------------------------------------------------------}
{                            F u n c t i o n s                                 }
{------------------------------------------------------------------------------}

function  TWXferMech      (hTmpWnd       : hWnd;
                           TransType     : TW_INT16) : TW_INT16;

function  TWAutofeed      (hTmpWnd       : hWnd)     : TW_INT16;

function  AutoFeedOn                     : bool;

function  TWAutofeedMenu  (hTmpWnd       : hWnd)     : TW_INT16;

{ MCM, FloatToFIX32 originally returns a TW_FIX32, however pascal cannot       }
{ handle this, therefore it was changed to pTW_FIX32.                          }

function  FloatToFIX32    (floater       : real)     : TW_FIX32;

function  FIX32ToFloat    (fix32         : TW_FIX32) : real;

procedure ShowImageInfo   (hWndofApp     : hWnd);

procedure ShowImageLayout (hWndofApp     : hWnd);

procedure ShowTW_ID       (hWndofApp     : hWnd;
                           twID          : TW_IDENTITY;
                           Text          : PChar);

procedure ShowTW_MSG      (hWndofApp     : hWnd;
                           lpMsg         : PMSG;
                           Text          : PChar);

procedure ShowCapability  (hWndofApp     : hWnd;
                           Cap           : TW_UINT16);

procedure GetItem         (Text          : PChar;
                           datatype      : TW_UINT16;
                           pData         : PChar);

procedure ShowPendingXfers(hWndofApp     : hWnd);

procedure ShowRC_CC       (hWndofApp     : hWnd;
                           DisplayRC     : TW_UINT16; { 0-No, 1-Yes.              }
                           ReturnCode    : TW_UINT16;
                           Dest          : TW_UINT16; { 0-Src Manager or 1-Source }
                           Text          : PChar;     { Additional Information    }
                           Title         : PChar);    { Triplet or Cap            }

function  MessageLevel                   : integer;

procedure SetMessageLevel (Level         : integer);

function  MsgGetCapability(Cap           : TW_UINT16;
                           Atype         : TW_UINT16;
                           ptwCapability : pTW_CAPABILITY) : BOOL;

function  MsgSetCapability(Cap           : TW_UINT16;
                           Atype         : TW_UINT16;
                           ptwCapability : pTW_CAPABILITY) : BOOL;


implementation

uses WinProcs, SysUtils, Dialogs,
     twa_glue, twa_acq, twa_apph, twa_spcl;

{------------------------------------------------------------------------------}
{ MCM.                                                                         }
{ CreateFileName - Generates a random file name for file transfer.             }
{ This procedure is not a part of the original TWAIN source code.              }
{                                                                              }
{------------------------------------------------------------------------------}

procedure CreateFileName(TmpName, TmpExt : PChar);
var i, j    : integer;
    TmpFile : File;
    TmpPath : array[0..1024] of char;
begin
  if (GetTempPath(SizeOf(TmpPath), TmpPath) = 0)
  then StrCopy(TmpPath, 'C:\');
  GetTempFileName(TmpPath, 'TWN', word(0), TmpName);
  Assign(TmpFile, TmpName);
  Reset(TmpFile);
  Close(TmpFile);
  Erase(TmpFile);
  i := 0;
  j := lstrlen(TmpName);
  while (TmpName[i] <> '.') and (i < j)
  do inc(i);
  TmpName[i+1] := #0;
  lstrcat(TmpName, TmpExt);
end; { End CreateFileName.                                                     }


{------------------------------------------------------------------------------}
{ TWXferMech - Set the current transfer mechanism for Twain based on the menus }
{ that have been checked by the user.                                          }
{                                                                              }
{ TransType = 0 Native Transfer.                                               }
{             1 Memory Transfer.                                               }
{             2 File Transfer.                                                 }
{------------------------------------------------------------------------------}

function TWXferMech(hTmpWnd   : hWnd;
                    TransType : TW_INT16) : TW_INT16;
var twRC    : TW_UINT16;
    Cap     : TW_CAPABILITY;
    pVal    : pTW_ONEVALUE;
{    ofn    : OPENFILENAME;} { MCM, Omitted since I don't like to assign       }
                             { file names just to do file transfer.            }
                             { I prefer that this is done automatically.       }
{    char *Filter[] =} { 'Bitmap(*.BMP)\0', '*.bmp\0\0'}{;}
  {$IFDEF VER70}
    Result  : TW_INT16;
  {$ENDIF}
    FileExt : array[0..4] of char; { MCM, added fore file type set-up.         }
begin
  LogMessage('TWXferMech');
  Cap.Cap     := ICAP_XFERMECH;
  Cap.ConType := TWON_ONEVALUE;

  { Alloc the container.                                                       }
  Cap.hContainer := GlobalAlloc(GHND, SizeOf(TW_ONEVALUE));
  if (Cap.hContainer <> 0)
  then begin
       pVal := pTW_ONEVALUE(GlobalLock(Cap.hContainer));
       pVal^.ItemType := TWTY_UINT16;

       case TransType of
       1 : begin { memory transfer.                                            }
             pVal^.Item := TWSX_MEMORY;
           end;
       2 : begin { file transfer.                                              }
             pVal^.Item := TWSX_FILE;
             lstrcpy(FileExt, 'BMP');

             { MCM Begin                                                   }
             { Check which file formats that are supported if any.         }
             { Mask out this part if you do not support BMP or TIFF.       }
             twRC := TWGetFileType;
             case twRC of
             TWFF_TIFF : lstrcpy(FileExt, 'TIF'); { Tag Image File Format. }
             TWFF_PICT : lstrcpy(FileExt, '');    { Macintosh PICT format. }
             TWFF_BMP  : lstrcpy(FileExt, 'BMP'); { Windows format.        }
             TWFF_XBM  : lstrcpy(FileExt, '');    { X-Windows Bitmap.      }
             TWFF_JFIF : lstrcpy(FileExt, '');    { JPEG File Format.      }
             else        begin
                           lstrcpy(FileExt, '');  { Unknown format.        }
                         end;
             end;
             CreateFileName(TWAINFileName, FileExt);
             { MCM, End.                                                   }

             { MCM, Omitted since I don't like to assign                   }
             { file names manually just to do file transfer.               }
             { I prefer that this is done automatically. (see above)       }

             {
             if (at.SaveAs[0] <> #0)
             then lstrcpyn(TWAINFilename, at.SaveAs, SizeOf(TWAINFilename));
             else begin} { file transfer currently fixed to bitmap format. }
                        { get the filename to save as                      }
                        { check formats supported by the source            }
             {
                  ofn.lStructSize       := SizeOf(OPENFILENAME);
                  ofn.hwndOwner         := hTmpWnd;
                  ofn.lpstrFilter       := Filter[0];
                  ofn.lpstrCustomFilter := Nil;
                  ofn.nMaxCustFilter    := 0;
                  ofn.nFilterIndex      := 0;
                  ofn.lpstrFile         := TWAINFilename;
                  ofn.nMaxFile          := SizeOf(TWAINFilename);
                  ofn.lpstrFileTitle    := Nil;
                  ofn.nMaxFileTitle     := 0L;
                  ofn.lpstrInitialDir   := Nil;
                  ofn.lpstrTitle        := Nil;
                  ofn.Flags             := 0;
                  ofn.nFileOffset       := 0;
                  ofn.nFileExtension    := 0;
                  ofn.lpstrDefExt       := 'BMP';
                  if (GetSaveFileName(@ofn) = False)
                  then Result := TWRC_FAILURE;
             end;
                  }
           end
       else begin { Native Tranfer.                                            }
            pVal^.Item := TWSX_NATIVE;
            end;
       end; { End Case                                                         }
       GlobalUnlock(Cap.hContainer);

       Result := lpDSM_Entry(@appID,
                             @dsID,
                             DG_CONTROL,
                             DAT_CAPABILITY,
                             MSG_SET,
                             TW_MEMREF(@Cap));
       GlobalFree(Cap.hContainer);
       if (Result <> TWRC_SUCCESS)
       then begin
            if (MessageLevel >= ML_ERROR)
            then ShowRC_CC(hTmpWnd, 1, Result, 1,
                           '',
                           'MSG_SET of ICAP_XFERMECH');
       end;
  end
  else begin
       Result := TWRC_FAILURE;
       if (MessageLevel >= ML_ERROR)
       then ShowRC_CC(hTmpWnd, 0, 0, 0,
                      'Memory Allocation Failed',
                      'MSG_SET of ICAP_XFERMECH');
  end;
  TWXferMech := Result;
end; { End TWXferMech.                                                         }


{------------------------------------------------------------------------------}
{ TWAutofeed - Checks if the autofeed option on the FILE menu is               }
{ checked and if so begins a capability negotiation.  If the ADF is already    }
{ enabled, (check this) the return value is ??? otherwise the application      }
{ negotiates with the source to enable the capability                          }
{                                                                              }
{ If the ADF at the source cannot be enabled, the application should post an   }
{ error and not allow a transfer (ie: the Source's UI to come up).             }
{------------------------------------------------------------------------------}

function TWAutofeed(hTmpWnd : hWnd) : TW_INT16;
var Cap     : TW_CAPABILITY;
    pVal    : pTW_ONEVALUE;
  {$IFDEF VER70}
    Result  : TW_INT16;
  {$ENDIF}
    Details : array[0..128] of char;
begin
  Result       := TWRC_SUCCESS;
  AutoFeedBOOL := False;
  StrCopy(Details, '');

  { If Automatic Feeding > CAP_AUTOFEED On is checked.                         }
  if (GetMenuState(GetMenu(hTmpWnd),
                   TW_APP_AUTOFEED,
                   MF_BYCOMMAND) = MF_CHECKED) or True
  then begin
       Cap.Cap     := CAP_FEEDERENABLED;
       Cap.ConType := TWON_ONEVALUE;

       Result := lpDSM_Entry(@appID,
                             @dsID,
                             DG_CONTROL,
                             DAT_CAPABILITY,
                             MSG_GET,
                             TW_MEMREF(@Cap));

       if (Result <> TWRC_SUCCESS)
       then begin
            Result := TWRC_FAILURE;
            lstrcpy(Details, 'CAP_FEEDERENABLED');
            GlobalFree(Cap.hContainer);
       end
       else begin
             pVal := pTW_ONEVALUE(GlobalLock(Cap.hContainer));
             if (pVal^.Item <> 0)
             then begin { Feeder is enabled no need to negotiate.              }
                  GlobalUnlock(Cap.hContainer);
                  GlobalFree(Cap.hContainer);
             end
             else begin
	          { Negotiate with the source.                                 }
                  { Try to turn on CAP_FEEDERENABLED.                          }
                  pVal^.ItemType := TWTY_BOOL;
                  pVal^.Item     := 1;
                  GlobalUnlock(Cap.hContainer);
                  Result := lpDSM_Entry(@appID,
                                        @dsID,
                                        DG_CONTROL,
                                        DAT_CAPABILITY,
                                        MSG_SET,
                                        TW_MEMREF(@cap));

                  { Free, because the GET call will allocate a new container.  }
                  GlobalFree(Cap.hContainer);

                  if (Result = TWRC_SUCCESS)
                  then begin { Verify that CAP_FEEDERENABLED is now TRUE.      }
                       Result := lpDSM_Entry(@appID,
                                             @dsID,
                                             DG_CONTROL,
                                             DAT_CAPABILITY,
                                             MSG_GET,
                                             TW_MEMREF(@Cap));

                       if (Result = TWRC_SUCCESS)
                       then begin
                            pVal := pTW_ONEVALUE(GlobalLock(Cap.hContainer));
                            if (pVal^.Item = 0{!= True})
                            then begin
                                 { Verification failed -                       }
                                 { CAP_FEEDERENABLED is FALSE                  }
                                 { even after successful set to TRUE.          }
                                 Result := TWRC_FAILURE;
                                 StrCopy(Details, 'CAP_FEEDERENABLED');
                            end;
                            GlobalUnlock(Cap.hContainer);
                            GlobalFree(Cap.hContainer);
                       end
                       else begin
                            Result := TWRC_FAILURE;
                            StrCopy(Details, 'CAP_FEEDERENABLED');
                       end;
                  end
                  else begin
                       { MSG_SET of CAP_FEEDERENABLED to TRUE did not succeed. }
                       StrCopy(Details, 'CAP_FEEDERENABLED');
                       Result := TWRC_FAILURE;
                  end;
             end;

             if (Result = TWRC_SUCCESS)
             then begin { CAP_AUTOFEED.                                        }
                  Cap.Cap     := CAP_AUTOFEED;
                  Cap.ConType := TWON_ONEVALUE;
                  Result := lpDSM_Entry(@appID,
                                        @dsID,
                                        DG_CONTROL,
                                        DAT_CAPABILITY,
                                        MSG_GET,
                                        TW_MEMREF(@Cap));

                  if (Result <> TWRC_SUCCESS)
                  then begin { MSG_GET on CAP_AUTOFEED did not succeed.        }
                      Result := TWRC_FAILURE;
                      StrCopy(Details, 'CAP_AUTOFEED');
                  end;

                  { MSG_GET on CAP_AUTOFEED returned success.                  }
                  pVal := pTW_ONEVALUE(GlobalLock(Cap.hContainer));
                  if (pVal^.Item <> 0)
                  then begin { CAP_AUTOFEED is on.                             }
                       GlobalUnlock(Cap.hContainer);
                       GlobalFree(Cap.hContainer);
                  end
                  else begin { Try to set CAP_AUTOFEED to TRUE.                }
                       pVal^.ItemType := TWTY_BOOL;
                       pVal^.Item     := 1;
                       GlobalUnlock(Cap.hContainer);
                       Result := lpDSM_Entry(@appID,
                                             @dsID,
                                             DG_CONTROL,
                                             DAT_CAPABILITY,
                                             MSG_SET,
                                             TW_MEMREF(@Cap));
                       GlobalFree(Cap.hContainer);

                       if (Result = TWRC_SUCCESS)
                       then begin { Verify that CAP_AUTOFEED is on.            }
                            Result := lpDSM_Entry(@appID,
                                                  @dsID,
                                                  DG_CONTROL,
                                                  DAT_CAPABILITY,
                                                  MSG_GET,
                                                  TW_MEMREF(@Cap));
                            if (Result = TWRC_SUCCESS)
                            then begin
                                 pVal := pTW_ONEVALUE(GlobalLock(Cap.hContainer));
                                 if (pVal^.Item = 0)
                                 then begin
                                      Result := TWRC_FAILURE;
                                      StrCopy(Details, 'CAP_AUTOFEED');
                                      { CAP_AUTOFEED returns FALSE even after  }
                                      { succesful set to TRUE                  }
                                 end;
                                 GlobalUnlock(Cap.hContainer);
                                 GlobalFree(Cap.hContainer);
                            end
                            else begin
                                 Result := TWRC_FAILURE;
                                 StrCopy(Details, 'CAP_AUTOFEED');
                            end;
                       end
                       else begin
                            { MSG_SET of CAP_AUTOFEED to TRUE did not succeed. }
                            Result := TWRC_FAILURE;
                            StrCopy(Details, 'CAP_AUTOFEED');
                       end;
                  end;
             end;
       end;

       { Set Local Autofeed Variable on/off.                                   }
       if (Result = TWRC_SUCCESS)
       then begin
            AutoFeedBOOL := True;
       end
       else begin
            AutoFeedBOOL := False;
            if (MessageLevel >= ML_ERROR)
            then ShowRC_CC(hTmpWnd, 0, 0, 0, Details, 'CAP_AUTOFEED ERROR');
       end;
  end;
  TWAutofeed := Result;
end; { End TWAutofeed.                                                         }


{------------------------------------------------------------------------------}
{ FUNCTION: AutoFeedOn                                                         }
{                                                                              }
{ ARGS:    none                                                                }
{                                                                              }
{ RETURNS: current state of Source Manager (open/closed)                       }
{                                                                              }
{ NOTES:   Just a way to reduce the number of global vars                      }
{------------------------------------------------------------------------------}

function AutoFeedOn : bool;
begin
  AutoFeedOn := AutoFeedBOOL;
end; { End AutoFeedOn.                                                         }


{------------------------------------------------------------------------------}
{ TWAutofeedMenu - Set the current Autofeed Options and Flags based on the     }
{ menus that have been checked by the user.                                    }
{------------------------------------------------------------------------------}

function TWAutofeedMenu(hTmpWnd : hWnd) : TW_INT16;
  {$IFDEF VER70}
var Result : TW_INT16;
  {$ENDIF}
begin
  Result := TWAutoFeed(hTmpWnd);
end; { End TWAutofeedMenu.                                                     }


{------------------------------------------------------------------------------}
{ FloatToFIX32 - Convert a floating point value into a FIX32.                  }
{------------------------------------------------------------------------------}

function FloatToFIX32(floater : real) : TW_FIX32;
var Fix32_value : TW_FIX32;
    Value       : TW_INT32;
begin
  Value := TW_INT32(Round(floater * 65536.0 + 0.50));
  Fix32_value.Whole := LOWORD(value shr 16);
  Fix32_value.Frac  := LOWORD(value and $0000FFFF);
  FloatToFIX32      := Fix32_value;
end; { End FloatToFIX32.                                                       }


{------------------------------------------------------------------------------}
{ FIX32ToFloat - Convert a FIX32 value into a floating point value             }
{------------------------------------------------------------------------------}

function FIX32ToFloat(fix32 : TW_FIX32) : real;
var floater : real;
begin
  floater      := 1.0 * fix32.Whole + (1.0 * fix32.Frac / 65536.0);
  FIX32ToFloat := floater;
end; { End FIX32ToFloat.                                                       }


{------------------------------------------------------------------------------}
{ ShowImageInfo -                                                              }
{------------------------------------------------------------------------------}

procedure ShowImageInfo(hWndofApp : hWnd);
var twRC        : TW_UINT16;
    ccRC        : TW_UINT16;
    twImageInfo : TW_IMAGEINFO;
    Details     : array[0..255] of char;
    Details2    : array[0..1024] of char;
    pTWFix32    : pTW_FIX32;
    index       : TW_UINT16;
    twStatus    : TW_STATUS;
    ValStr      : array[0..255] of char;
begin
  { Check ImageInfo information.                                               }
  twRC := lpDSM_Entry(@appID,
                      @dsID,
                      DG_IMAGE,
                      DAT_IMAGEINFO,
                      MSG_GET,
                      TW_MEMREF(@twImageInfo));

  { Return Code from Operation.                                                }
  StrFmt(Details2, 'RETURN CODE: %d', [twRC]);

  case twRC of
  TWRC_SUCCESS : begin
                   StrCat(Details2, CR_LF);
                   str(FIX32ToFloat(twImageInfo.XResolution):4:4, ValStr);
                   StrCat(Details2, 'XResolution: ');
                   StrCat(Details2, ValStr);

                   StrCat(Details2, CR_LF);
                   pTWFix32 := @(twImageInfo.YResolution);
                   str(FIX32ToFloat(pTWFix32^):4:4, ValStr);
                   StrCat(Details2, 'YResolution: ');
                   StrCat(Details2, ValStr);

                   StrCat(Details2, CR_LF);
                   StrFmt(Details, 'ImageWidth: %d', [twImageInfo.ImageWidth]);
                   StrCat(Details2, Details);

                   StrCat(Details2, CR_LF);
                   StrFmt(Details, 'ImageLength: %d', [twImageInfo.ImageLength]);
                   StrCat(Details2, Details);

                   StrCat(Details2, CR_LF);
                   StrFmt(Details, 'SamplesPerPixel: %d', [twImageInfo.SamplesPerPixel]);
                   StrCat(Details2, Details);

                   StrCat(Details2, CR_LF);
                   StrFmt(Details, 'BitsPerSample: %d', [twImageInfo.BitsPerSample[0]]);
                   StrCat(Details2, Details);
                   for index := 1 to 7
                   do begin
                      StrFmt(Details, ', %d', [twImageInfo.BitsPerSample[index]]);
                      StrCat(Details2, Details);
                   end;

                   StrCat(Details2, CR_LF);
                   StrFmt(Details, 'BitsPerPixel: %d', [twImageInfo.BitsPerPixel]);
                   StrCat(Details2, Details);

                   StrCat(Details2, CR_LF);
                   StrFmt(Details, 'Planar: %d', [twImageInfo.Planar]);
                   StrCat(Details2, Details);

                   StrCat(Details2, CR_LF);
                   StrFmt(Details, 'PixelType: %d', [twImageInfo.PixelType]);
                   StrCat(Details2, Details);

                   StrCat(Details2, CR_LF);
                   StrFmt(Details, 'Compression: %d', [twImageInfo.Compression]);
                   StrCat(Details2, Details);
                 end;
{  TWRC_FAILURE : begin
                 end; }
  else           begin { Determine details of failure from Source.             }
                   ccRC := lpDSM_Entry(@appID,
                                       @dsID,
                                       DG_CONTROL,
                                       DAT_STATUS,
                                       MSG_GET,
                                       TW_MEMREF(@twStatus));

                   if (ccRC = TWRC_SUCCESS)
                   then begin
                        StrCat(Details2, CR_LF);
                        StrFmt(Details, 'Condition Code: %d', [twStatus.ConditionCode]);
                        StrCat(Details2, Details);
                   end
                   else begin
                        StrCat(Details2, CR_LF);
                        StrCopy(Details, 'Condition Code Check Failed');
                        StrCat(Details2, Details);
                   end;
                 end;
  end;
  ShowRC_CC(hWndofApp, 0, 0, 0, Details2, 'TW_IMAGEINFO');
end; { End ShowImageInfo.                                                      }


{------------------------------------------------------------------------------}
{ ShowImageLayout                                                              }
{------------------------------------------------------------------------------}

procedure ShowImageLayout(hWndofApp : hWnd);
var twRC          : TW_UINT16;
    ccRC          : TW_UINT16;
    twImageLayout : TW_IMAGELAYOUT;
    Details       : array[0..255] of char;
    Details2      : array[0..1024] of char;
    twStatus      : TW_STATUS;
    ValStr        : array[0..255] of char;
begin
  { Check ImageLayout information.                                             }
  twRC := lpDSM_Entry(@appID,
                      @dsID,
                      DG_IMAGE,
                      DAT_IMAGELAYOUT,
                      MSG_GET,
                      TW_MEMREF(@twImageLayout));

  { Return Code from Operation.                                                }
  StrFmt(Details2, 'RETURN CODE: %d', [twRC]);

  case twRC of
  TWRC_SUCCESS : begin
                   StrCat(Details2, CR_LF);
                   Str(FIX32ToFloat(twImageLayout.Frame.Left):4:4, ValStr);
                   StrCat(Details2, 'Frame.Left: ');
                   StrCat(Details2, ValStr);

                   StrCat(Details2, CR_LF);
                   Str(FIX32ToFloat(twImageLayout.Frame.Top):4:4, ValStr);
                   StrCat(Details2, 'Frame.Top: ');
                   StrCat(Details2, ValStr);

                   StrCat(Details2, CR_LF);
                   Str(FIX32ToFloat(twImageLayout.Frame.Right):4:4, ValStr);
                   StrCat(Details2, 'Frame.Right: ');
                   StrCat(Details2, ValStr);

                   StrCat(Details2, CR_LF);
                   Str(FIX32ToFloat(twImageLayout.Frame.Bottom):4:4, ValStr);
                   StrCat(Details2, 'Frame.Bottom: ');
                   StrCat(Details2, ValStr);

                   StrCat(Details2, CR_LF);
                   StrFmt(Details, 'DocumentNumber: %d', [twImageLayout.DocumentNumber]);
                   StrCat(Details2, Details);

                   StrCat(Details2, CR_LF);
                   StrFmt(Details, 'PageNumber: %d', [twImageLayout.PageNumber]);
                   StrCat(Details2, Details);

                   StrCat(Details2, CR_LF);
                   StrFmt(Details, 'FrameNumber: %d', [twImageLayout.FrameNumber]);
                   StrCat(Details2, Details);
                 end;
{  TWRC_FAILURE : begin
                 end;  }
  else           begin { determine details of failure from Source.             }
                   ccRC:= lpDSM_Entry(@appID,
                                      @dsID,
                                      DG_CONTROL,
                                      DAT_STATUS,
                                      MSG_GET,
                                      TW_MEMREF(@twStatus));

                   if (ccRC = TWRC_SUCCESS)
                   then begin
                        StrCat(Details2, CR_LF);
                        StrFmt(Details, 'Condition Code: %d', [twStatus.ConditionCode]);
                        StrCat(Details2, Details);
                   end
                   else begin
                        StrCat(Details2, CR_LF);
                        StrCat(Details2, 'Condition Code Check Failed');
                   end;
                 end;
  end;
  ShowRC_CC(hWndofApp, 0, 0, 0,Details2, 'TW_IMAGELAYOUT');
end; { End ShowImageLayout.                                                    }


{------------------------------------------------------------------------------}
{ ShowTW_ID -                                                                  }
{------------------------------------------------------------------------------}

procedure ShowTW_ID(hWndofApp : hWnd; twID : TW_IDENTITY; Text : PChar);
var Details  : array[0..255] of char;
    Details2 : array[0..1024] of char;
begin
  StrFmt(Details2, 'Id: %d', [twID.Id]);
  StrCat(Details2, CR_LF);
  StrFmt(Details, 'TW_VERSION.MajorNum: %d', [twID.Version.MajorNum]);
  StrCat(Details2, Details);
  StrCat(Details2, CR_LF);
  StrFmt(Details, 'TW_VERSION.MinorNum: %d', [twID.Version.MinorNum]);
  StrCat(Details2, Details);
  StrCat(Details2, CR_LF);
  StrFmt(Details, 'TW_VERSION.Language: %d', [twID.Version.Language]);
  StrCat(Details2, Details);
  StrCat(Details2, CR_LF);
  StrFmt(Details, 'TW_VERSION.Country: %d', [twID.Version.Country]);
  StrCat(Details2, Details);
  StrCat(Details2, CR_LF);
  StrCat(Details2, 'TW_VERSION.Info: ');
  StrCat(Details2, twID.Version.Info);
  StrCat(Details2, CR_LF);
  StrFmt(Details, 'ProtocolMajor: %d', [twID.ProtocolMajor]);
  StrCat(Details2, Details);
  StrCat(Details2, CR_LF);
  StrFmt(Details, 'ProtocolMinor: %d', [twID.ProtocolMinor]);
  StrCat(Details2, Details);
  StrCat(Details2, CR_LF);
  StrFmt(Details, 'SupportedGroups: %d', [twID.SupportedGroups]);
  StrCat(Details2, Details);
  StrCat(Details2, CR_LF);
  StrCat(Details2, 'Manufacturer: ');
  StrCat(Details2, twID.Manufacturer);
  StrCat(Details2, CR_LF);
  StrCat(Details2, 'ProductFamily: ');
  StrCat(Details2, twID.ProductFamily);
  StrCat(Details2, CR_LF);
  StrCat(Details2, 'ProductName: ');
  StrCat(Details2, twID.ProductName);
  StrCat(Details2, CR_LF);

  ShowRC_CC(hWndofApp, 0, 0, 0, Details2, Text);
end; { End ShowTW_ID.                                                          }


{------------------------------------------------------------------------------}
{ ShowTW_MSG - MCM, Not part of TWAIN toolkit, use with care !!!               }
{------------------------------------------------------------------------------}

procedure ShowTW_MSG(hWndofApp : hWnd; lpMsg : PMSG; Text : PChar);
var Details  : array[0..255] of char;
    Details2 : array[0..1024] of char;
begin
  if (GetKeyState(vk_Shift) < 0) { Shift - key is up.                          }
  then begin
       if (lpMsg <> Nil)
       then begin
            StrCopy(Details2, 'lpMsg / TMessage (App.)');
            StrCat(Details2, CR_LF);
            StrFmt(Details, 'hWnd:    %u', [lpMsg^.hwnd]);
            StrCat(Details2, Details);
            StrCat(Details2, CR_LF);
            StrFmt(Details, 'Message: %u', [lpMsg^.message]);
            StrCat(Details2, Details);
            StrCat(Details2, CR_LF);
            StrFmt(Details, 'wParam:  %u', [lpMsg^.wParam]);
            StrCat(Details2, Details);
            StrCat(Details2, CR_LF);
            StrFmt(Details, 'lParam:  %ld', [lpMsg^.lParam]);
            StrCat(Details2, Details);
            StrCat(Details2, CR_LF);
            StrFmt(Details, 'Time:    %ld', [lpMsg^.time]);
            StrCat(Details2, Details);
            StrCat(Details2, CR_LF);
            StrFmt(Details, 'pt.x:    %u', [lpMsg^.pt.x]);
            StrCat(Details2, Details);
            StrCat(Details2, CR_LF);
            StrFmt(Details, 'pt.y:    %u', [lpMsg^.pt.y]);
            StrCat(Details2, Details);
            StrCat(Details2, CR_LF);

            MessageBox(hWndofApp, Details2, Text, MB_OK);
       end;
  end;
end; { End ShowTW_MSG.                                                         }


{------------------------------------------------------------------------------}
{ ShowCapability -                                                             }
{------------------------------------------------------------------------------}

procedure ShowCapability(hWndofApp : hWnd; Cap : TW_UINT16);
var twRC         : TW_UINT16;
    ccRC         : TW_UINT16;
    twCapability : TW_CAPABILITY;
    Details      : array[0..255] of char;
    Details2     : array[0..1024] of char;
    index        : TW_UINT16;
    twStatus     : TW_STATUS;
    pEnum        : pTW_ENUMERATION;
    pOneV        : pTW_ONEVALUE;
    pRange       : pTW_RANGE;
    pArray       : pTW_ARRAY;
    pData        : PChar;
    datasize     : TW_INT16;
begin
  twCapability.Cap     := Cap;
  twCapability.ConType := TWON_DONTCARE16;

  { Check Capability information.                                              }
  twRC := lpDSM_Entry(@appID,
                      @dsID,
                      DG_CONTROL,
                      DAT_CAPABILITY,
                      MSG_GET,
                      TW_MEMREF(@twCapability));

  { Return Code from Operation.                                                }
  StrFmt(Details, 'RETURN CODE: %d', [twRC]);
  StrCopy(Details2, Details);

  { Show TW_CAPABILITY Contents.                                               }
  MatchTwainInt('CAP', twCapability.Cap, Details, SizeOf(Details));
  StrCat(Details2, CR_LF);
  StrCat(Details2, 'Cap: ');
  StrCat(Details2, Details);

  case twRC of
  TWRC_SUCCESS : begin
                   MatchTwainInt('ConType', twCapability.ConType, Details, SizeOf(Details));
                   StrCat(Details2, CR_LF);
                   StrCat(Details2, 'ConType: ');
                   StrCat(Details2, Details);

                   { Show Container Contents.                                  }
                   StrCat(Details2, CR_LF);
                   StrCat(Details2, 'hContainer.');
                   case twCapability.ConType of
                   TWON_ONEVALUE    : if (twCapability.hContainer <> 0)
                                      then begin
                                           pOneV := pTW_ONEVALUE(GlobalLock(twCapability.hContainer));
                                           StrCat(Details2, CR_LF);
                                           StrCat(Details2, '     ItemType: ');
                                           MatchTwainInt('ItemType', pOneV^.ItemType, Details, SizeOf(Details));
                                           StrCat(Details2, Details);

                                           StrCat(Details2, CR_LF);
                                           StrCat(Details2, '     Item: ');
                                           GetItem(Details, pOneV^.ItemType, PChar(@pOneV^.Item));
                                           StrCat(Details2, Details);

                                           GlobalUnlock(twCapability.hContainer);
                                           GlobalFree(twCapability.hContainer);
                                      end
                                      else MessageBox(0, 'TWON_ONEVALUE, twCapability.hContainer = 0',
                                                         'ShowCapability Error', MB_OK);
                   TWON_ENUMERATION : if (twCapability.hContainer <> 0)
                                      then begin
                                           pEnum := pTW_ENUMERATION(GlobalLock(twCapability.hContainer));
                                           StrCat(Details2, CR_LF);
                                           StrCat(Details2, '     ItemType: ');
                                           MatchTwainInt('ItemType', pEnum^.ItemType, Details, SizeOf(Details));
                                           datasize:= GetPrivateProfileInt('ItemSize', Details, 4, IniFile);
                                           StrCat(Details2, Details);

                                           StrCat(Details2, CR_LF);
                                           StrFmt(Details, '     NumItems: %d', [pEnum^.NumItems]);
                                           StrCat(Details2, Details);

                                           StrCat(Details2, CR_LF);
                                           StrFmt(Details, '     CurrentIndex: %d', [pEnum^.CurrentIndex]);
                                           StrCat(Details2, Details);

                                           StrCat(Details2, CR_LF);
                                           StrFmt(Details, '     DefaultIndex: %d', [pEnum^.DefaultIndex]);
                                           StrCat(Details2, Details);

                                           StrCat(Details2, CR_LF);
                                           StrCat(Details2, '     ItemList: ');
                                           pData := @pEnum^.ItemList;

                                           { index = 0.                           }
                                           GetItem(Details, pEnum^.ItemType, pData);
                                           StrCat(Details2, Details);
                                           pData := pData + datasize;
                                           for index := 1 to (pEnum^.NumItems - 1)
                                           do begin
                                              StrCat(Details2, ', ');
                                              GetItem(Details, pEnum^.ItemType, pData);
                                              StrCat(Details2, Details);
                                              pData := pData + datasize;
                                           end;
                                           GlobalFree(twCapability.hContainer);
                                      end
                                      else MessageBox(0, 'TWON_ENUMERATION, twCapability.hContainer = 0',
                                                         'ShowCapability Error', MB_OK);
                   TWON_RANGE       : if (twCapability.hContainer <> 0)
                                      then begin
                                           pRange := pTW_RANGE(GlobalLock(twCapability.hContainer));

                                           StrCat(Details2, CR_LF);
                                           StrCat(Details2, '     ItemType: ');
                                           MatchTwainInt('ItemType', pRange^.ItemType, Details, SizeOf(Details));
                                           StrCat(Details2, Details);

                                           StrCat(Details2, CR_LF);
                                           StrCat(Details2, '     MinValue: ');
                                           GetItem(Details, pRange^.ItemType, PChar(@pRange^.MinValue));
                                           StrCat(Details2, Details);

                                           StrCat(Details2, CR_LF);
                                           StrCat(Details2, '     MaxValue: ');
                                           GetItem(Details, pRange^.ItemType, PChar(@pRange^.MaxValue));
                                           StrCat(Details2, Details);

                                           StrCat(Details2, CR_LF);
                                           StrCat(Details2, '     StepSize: ');
                                           GetItem(Details, pRange^.ItemType, PChar(@pRange^.StepSize));
                                           StrCat(Details2, Details);

                                           StrCat(Details2, CR_LF);
                                           StrCat(Details2, '     DefaultValue: ');
                                           GetItem(Details, pRange^.ItemType, PChar(@pRange^.DefaultValue));
                                           StrCat(Details2, Details);

                                           StrCat(Details2, CR_LF);
                                           StrCat(Details2, '     CurrentValue: ');
                                           GetItem(Details, pRange^.ItemType, PChar(@pRange^.CurrentValue));
                                           StrCat(Details2, Details);

                                           GlobalFree(twCapability.hContainer); { bad. }
                                      end
                                      else MessageBox(0, 'TWON_RANGE, twCapability.hContainer = 0',
                                                         'ShowCapability Error', MB_OK);
                   TWON_ARRAY       : if (twCapability.hContainer <> 0)
                                      then begin
                                           pArray := GlobalLock(twCapability.hContainer);

                                           StrCat(Details2, CR_LF);
                                           StrCat(Details2, '     ItemType: ');
                                           MatchTwainInt('ItemType', pArray^.ItemType, Details, SizeOf(Details));
                                           datasize:= GetPrivateProfileInt('ItemSize', Details, 4, IniFile);
                                           StrCat(Details2, Details);

                                           StrCat(Details2, CR_LF);
                                           StrFmt(Details, '     NumItems: %d', [pArray^.NumItems]);
                                           StrCat(Details2, Details);

                                           StrCat(Details2, CR_LF);
                                           StrCat(Details2, '     ItemList: ');
                                           pData := @pArray^.ItemList;

                                           { index = 0                              }
                                           GetItem(Details, pArray^.ItemType, pData);
                                           StrCat(Details2, Details);
                                           pData := pData + datasize;
                                           for index := 1 to (pArray^.NumItems - 1)
                                           do begin
                                              StrCat(Details2, ', ');
                                              GetItem(Details, pArray^.ItemType, pData);
                                              StrCat(Details2, Details);
                                              pData := pData + datasize;
                                           end;
                                           GlobalFree(twCapability.hContainer); { bad }
                                      end
                                      else MessageBox(0, 'TWON_ARRAY, twCapability.hContainer = 0',
                                                         'ShowCapability Error', MB_OK);
                   end;
                 end;
{  TWRC_FAILURE : begin
                 end; }
  else           begin { determine details of failure from Source.             }
                   ccRC := lpDSM_Entry(@appID,
                                       @dsID,
                                       DG_CONTROL,
                                       DAT_STATUS,
                                       MSG_GET,
                                       TW_MEMREF(@twStatus));

                   if (ccRC = TWRC_SUCCESS)
                   then begin
                        StrCat(Details2, CR_LF);
                        StrFmt(Details, 'Condition Code: %d', [twStatus.ConditionCode]);
                        StrCat(Details2, Details);
                   end
                   else begin
                        StrCat(Details2, CR_LF);
                        StrFmt(Details, 'Condition Code: %d', [twStatus.ConditionCode]);
                        StrCat(Details2, Details);
                   end;
                 end;
  end;
  ShowRC_CC(hWndofApp, 0, 0, 0, Details2, 'TW_CAPABILITY');
end; { End ShowCapability.                                                     }


{------------------------------------------------------------------------------}
{ GetItem -                                                                    }
{------------------------------------------------------------------------------}

procedure GetItem(Text : Pchar; datatype : TW_UINT16; pData : PChar);
var twInt8   : TW_INT8;
    twInt16  : TW_INT16;
    twInt32  : TW_INT32;
    twFix32  : TW_FIX32;
    twUInt32 : TW_UINT32;
    Value    : real;
    TmpText  : array[0..64] of char;
begin
  case datatype of
  0  : begin  { TWTY_INT8                                                      }
         twInt8 := TW_INT8(pData^);
         Str(twInt8, TmpText);
       end;
  3  : begin { TWTY_UINT8                                                      }
         twInt8 := TW_INT8(pData^);
         Str(twInt8, TmpText);
       end;
  1,         { TWTY_INT16                                                      }
  44 : begin { TWTY_HANDLE                                                     }
         twInt16 := TW_INT16(pData^);
         Str(twInt16, TmpText);
       end;
  4,         { TWTY_UINT16                                                     }
  6  : begin { TWTY_BOOL                                                       }
         twInt16 := TW_UINT16(pData^);
         Str(twInt16, TmpText);
       end;
  2  : begin { TWTY_INT32                                                      }
         twInt32 := TW_INT32(pData^);
         Str(twInt32, TmpText);
       end;
  5,         { TWTY_UINT32                                                     }
  43 : begin { TWTY_MEMREF                                                     }
         twUInt32 := TW_UINT32(pData^);
         Str(twUInt32, TmpText);
       end;
  7  : begin { TWTY_FIX32                                                      }
         twFix32.Whole := pTW_FIX32(pData)^.Whole;
         twFix32.Frac  := pTW_FIX32(pData)^.Frac;
         value         := FIX32ToFloat(twFix32);
         Str(value:4:4, TmpText);
       end;
 { 9  : } { TWTY_STR32                                                         }
 { 10 : } { TWTY_STR64                                                         }
 { 11 : } { TWTY_STR128                                                        }
 { 12 : } { TWTY_STR255                                                        }
  else begin
         lstrcpy(Text, pData);
       end;
  end;
  StrCopy(Text, TmpText);
end; { End GetItem.                                                            }


{------------------------------------------------------------------------------}
{ ShowPendingXfers -                                                           }
{------------------------------------------------------------------------------}

procedure ShowPendingXfers(hWndofApp : hWnd);
var twRC           : TW_UINT16;
    ccRC           : TW_UINT16;
    twPendingXfers : TW_PENDINGXFERS;
    Details        : array[0..255] of char;
    Details2       : array[0..1024] of char;
    twStatus       : TW_STATUS;
begin
  { Inquire Pending Xfers.                                                     }
  twRC := lpDSM_Entry(@appID,
                      @dsID,
                      DG_CONTROL,
                      DAT_PENDINGXFERS,
                      MSG_GET,
                      TW_MEMREF(@twPendingXfers));

  { Return Code from Operation.                                                }
  StrCat(Details2, CR_LF);
  StrFmt(Details, 'RETURN CODE: %d', [twRC]);

  StrCat(Details2, CR_LF);
  case twRC of
  TWRC_SUCCESS : begin
                   StrFmt(Details, 'Count: %d', [twPendingXfers.Count]);
                   StrCat(Details2, Details);
                 end;
{ TWRC_FAILURE : begin
                 end; }
  else           begin { determine details of failure from SM.                 }
                   ccRC := lpDSM_Entry(@appID,
                                       @dsID,
                                       DG_CONTROL,
                                       DAT_STATUS,
                                       MSG_GET,
                                       TW_MEMREF(@twStatus));

                   if (ccRC = TWRC_SUCCESS)
                   then begin
                        StrFmt(Details, 'Condition Code:  %d', [twStatus.ConditionCode]);
                        StrCat(Details2, Details);
                   end
                   else begin
                        StrFmt(Details, 'Condition Code:  %d', [twStatus.ConditionCode]);
                        StrCat(Details2, Details);
                   end;
                 end;
  end;
  ShowRC_CC(hWndofApp, 0, 0, 0, Details2, 'ADF Pending Transfers');
end; { End ShowPendingXfers.                                                   }


{------------------------------------------------------------------------------}
{ ShowRC_CC - Main message notification routine.                               }
{ Displays the failure, condition codes and message associated with the        }
{ problem. Additionally, all log messages pass thru this function.             }
{------------------------------------------------------------------------------}

procedure ShowRC_CC(hWndofApp  : hWnd;
                    DisplayRC  : TW_UINT16; { 0-No, 1-Yes.                     }
                    ReturnCode : TW_UINT16;
                    Dest       : TW_UINT16; { 0-Source Manager or 1-Source     }
                    Text       : PChar;     { Additional Information           }
                    Title      : PChar);    { Triplet or Cap                   }
var twRC      : TW_UINT16;
    Details   : array[0..255] of char;
    Details2  : array[0..1024] of char;
    twStatus  : TW_STATUS;
    outstring : array[0..1024] of char;
    outtitle  : array[0..255] of char;
begin
  StrCopy(Details,  '');
  StrCopy(Details2, '');

  lstrcpy(outstring, Text);
  lstrcpy(outtitle,  Title);

  { if display check and display condition code.                               }
  if (DisplayRC = 1)
  then begin
       {Get Return Code to be displayed, place in Details2.                    }
       MatchTwainInt('ReturnCode', TW_INT32(ReturnCode),
                     Details2,     SizeOf(Details2));

       if (ReturnCode <> TWRC_SUCCESS)
       then begin
            if (Dest = 0) {Source Manager.                                     }
            then begin
                 if TWIsDSMOpen { The Source Manager is open.                  }
                 then begin
                      twRC := lpDSM_Entry(@appID,
                                          Nil,
                                          DG_CONTROL,
                                          DAT_STATUS,
                                          MSG_GET,
                                          TW_MEMREF(@twStatus));
                 end
                 else begin { source mgr not open.                             }
                      twRC := TWRC_FAILURE;
                 end;
            end
            else begin { Check that a DS is open.                              }
                 if TWIsDSOpen
                 then begin
                      twRC := lpDSM_Entry(@appID,
                                          @dsID,
                                          DG_CONTROL,
                                          DAT_STATUS,
                                          MSG_GET,
                                          TW_MEMREF(@twStatus));
                 end
                 else begin { Source is not open.                              }
                      twRC := TWRC_FAILURE;
                 end;
            end;
       end
       else twRC := TWRC_FAILURE;
       if (twRC = TWRC_SUCCESS)
       then begin
            MatchTwainInt('ConditionCode', TW_INT32(twStatus.ConditionCode),
                          Details, SizeOf(Details));
       end
       else begin
            StrCopy(Details, 'Unable to determine Condition Code of Failure');
       end;

       lstrcpy(outtitle, 'Operation Status');
       lstrcpy(outstring, 'Operation: ');
       lStrCat(outstring, Title);
       lStrCat(outstring, CR_LF);
       lStrCat(outstring, 'Return Code: ');
       lStrCat(outstring, Details2);
       lStrCat(outstring, CR_LF);
       lStrCat(outstring, 'Condition Code: ');
       lStrCat(outstring, Details);
  end;

  if True {(TagATReg.AutotestMode = ATTENDED)}
  then MessageBox(hWndofApp, outstring, outtitle, MB_OK);
{ MCM, ? else TimedMsgBox(hInst, hWndofApp, outstring, outtitle, MB_OK, IDOK, TimeOut); }

{  if (TagATReg.Log = ON)
  then begin
       lStrCat(outtitle, '\r\n');
       lStrCat(outtitle, outstring);
       lStrCat(outtitle, '\r\n');
       LogOut(outtitle,  lstrlen(outtitle));
  end;
}
end; { End ShowRC_CC.                                                          }


{------------------------------------------------------------------------------}
{ FUNCTION: MessageLevel                                                       }
{                                                                              }
{ ARGS:    none                                                                }
{                                                                              }
{ RETURNS: current state of Source Manager (open/closed)                       }
{                                                                              }
{ NOTES:   Just a way to reduce the number of global vars                      }
{------------------------------------------------------------------------------}

function MessageLevel : integer;
begin
  MessageLevel := MessageLevelVAR;
end; { End MessageLevel.                                                       }


{------------------------------------------------------------------------------}
{ FUNCTION: SetMessageLevel                                                    }
{                                                                              }
{ ARGS:    none                                                                }
{                                                                              }
{ RETURNS: current state of Source Manager (open/closed)                       }
{                                                                              }
{ NOTES:   Just a way to reduce the number of global vars                      }
{------------------------------------------------------------------------------}

procedure SetMessageLevel(Level : integer);
begin
  MessageLevelVAR := Level;
end; { End SetMessageLevel.                                                    }


{------------------------------------------------------------------------------}
{ MsgGetCapability - Sends one of the three GET triplets to the                }
{ source and returns the TW_CAPABILITY structure to the caller as a            }
{ parameter.  If the DSM_Entry call fails, FALSE is returned                   }
{------------------------------------------------------------------------------}

function MsgGetCapability(Cap           : TW_UINT16;
                          Atype         : TW_UINT16;
                          ptwCapability : pTW_CAPABILITY) : BOOL;
var rc     : TW_INT16;
    Text   : array[0..50] of char;
  {$IFDEF VER70}
    Result : BOOL;
  {$ENDIF}
begin {	Setup dsID for default Source.                                         }
  Result := True;
  if Not(TWIsDSOpen)
  then begin
       dsID.Id             := 0;
       dsID.ProductName[0] := #0;
       appID.Id            := 0; { Why do this?                                }
  end;

  { Open DSM.                                                                  }
  if Not(TWOpenDSM)
  then begin
       if (MessageLevel >= ML_ERROR)
       then ShowRC_CC(0, 0, 0, 0, 'Failed to Open DSM', 'Capabilities');
       Result := False;
  end;

  { Open DS.                                                                   }
  if Result
  then begin
       if Not(TWOpenDS)
       then begin
            if (MessageLevel >= ML_ERROR)
            then ShowRC_CC(0, 0, 0, 0, 'Failed to open DS', 'Capabilities');
            Result := False;
       end;
  end;

  if Result
  then begin
       ptwCapability^.Cap        := cap;
       ptwCapability^.ConType    := TWON_DONTCARE16;
       ptwCapability^.hContainer := 0;

       rc := lpDSM_Entry(@appID,
                         @dsID,
                         DG_CONTROL,
                         DAT_CAPABILITY,
                         AType,
                         TW_MEMREF(ptwCapability));

       if (rc = TWRC_SUCCESS)
       then Result := True
       else begin
            FillChar(Text, SizeOf(Text), 0);
            lStrCat(Text, 'DG_CONTROL / DAT_CAPABILITY / ');
            case AType of
            MSG_GET        : lStrCat(Text,'MSG_GET');
            MSG_GETDEFAULT : lStrCat(Text,'MSG_GETDEFAULT');
            MSG_GETCURRENT : lStrCat(Text,'MSG_GETCURRENT');
            end;

            { Error message using condition state of the Source.               }
            ShowRC_CC(0, 1, rc, 1, '', Text);
            Result := False;
       end;
  end;
  MsgGetCapability := Result;
end; { End MsgGetCapability.                                                   }


{------------------------------------------------------------------------------}
{ MsgSetCapability - Sends one of the two SET triplets to the source and       }
{ returns the TW_CAPABILITY structure to the caller as a parameter.            }
{ If the DSM_Entry call fails, FALSE is returned.                              }
{------------------------------------------------------------------------------}

function MsgSetCapability(Cap           : TW_UINT16;
                          Atype         : TW_UINT16;
                          ptwCapability : pTW_CAPABILITY) : BOOL;
var rc     : TW_INT16;
    Text   : array[0..50] of char;
  {$IFDEF VER70}
    Result : BOOL;
  {$ENDIF}
begin
  Result := True;
  { Setup dsID for default Source.                                             }
  if Not(TWIsDSOpen)
  then begin
       dsID.Id             := 0;
       dsID.ProductName[0] := #0;
       appID.Id            := 0; { Why do this?                                }
  end;

  { Open DSM.                                                                  }
  if (TWOpenDSM <> TRUE)
  then begin
       if (MessageLevel >= ML_ERROR)
       then ShowRC_CC(0, 0, 0, 0,'Failed to Open DSM','Capabilities');
       Result := False;
  end;

  { Open DS.                                                                   }
  if Result
  then begin
       if (TWOpenDS <> TRUE)
       then begin
            if (MessageLevel >= ML_ERROR)
            then ShowRC_CC(0, 0, 0, 0, 'Failed to open DS', 'Capabilities');
            Result := False;
       end;
  end;

  if Result
  then begin
       if (AType = MSG_RESET)
       then begin
            ptwCapability^.Cap        := Cap;
            ptwCapability^.ConType    := TWON_DONTCARE16;
            ptwCapability^.hContainer := 0;
       end;

       rc := lpDSM_Entry(@appID,
                         @dsID,
                         DG_CONTROL,
                         DAT_CAPABILITY,
                         AType,
                         TW_MEMREF(ptwCapability));

       if (rc = TWRC_SUCCESS)
       then Result := True
       else begin
            FillChar(Text,  SizeOf(Text), 0);
            lStrCat(Text, 'DG_CONTROL / DAT_CAPABILITY / ');
            case AType of
            MSG_SET   : lStrCat(Text, 'MSG_SET');
            MSG_RESET : lStrCat(Text, 'MSG_RESET');
            end;
            { Error message using condition state of the Source.               }
            ShowRC_CC(0, 1, rc, 1, '', Text);
            Result := False;
       end;
  end;
  MsgSetCapability := Result;
end; { End MsgSetCapability.                                                   }


end. { End Unit twa_capt.                                                      }
